#include <stdio.h>

/* Sample source code for TSMASM :

;                    N F F F F F F    
;              T   T O                
;   P T      N T A R D     R   J   S S
; T C R      T R L A T O L I   S   P P
; S Z A      S A U P R R E G A T O A A
; E I C P    E C R O A E F H L A P R R
; Q P E C    Q E S N P G T T U T C E E
; + - + +    + - + + + - - - - - - + +
  0 1 0 XX   0 1 0 0 0 1 1 1 1 1 1 0 0  ;		NORMAL CODE
  0 0 X 00   0 1 1 0 0 0 0 0 0 0 1 0 0	;(G0 = OREG)	TRAP ENTRY POINT

*/

char fname_v[] = {"TSM0.VLA"};
char fname_d[] = {"tsm0.hex"};
char xtab[] = {"0123456789ABCDEF"};

char prom[2][2048],p[2];

int prominit[2] = {0xb0, 0x00}; /* PROM initialization values khh 5/24/88 */

char	tseq[2],pczip[2],trace[2],pch[2],pcl[2];

int	ntseq,ttrace,alurs,trapon,
	nodtrap,f_oreg,f_left,f_right,f_alu,f_jstat,
	f_opc,spare1,spare2;

int jmask,jval;

main(argc,argv)
int argc;
char *argv[];
{
	FILE *fopen(),*f;
	char iline[256];
	int ilen,lnum,k,byte,aa,sum;
	register int i,j;

	printf("\n*** Falcon Trap State Machine Assembler\n");

	for(i=0; i<2048; i++)
		for(j=0; j<2; j++)
			prom[j][i] = 0;

	if(argc > 1)
		f = fopen(argv[1],"r");
	else
		f = fopen("TSM","r");

	if(f == 0) {
		printf("*** Can't open source file ***");
		exit();
	}
	else printf("*** Source file opened ....\n");

	lnum = 0;
	while(fgets(iline,256,f) != NULL) {
		lnum += 1;
		printf("\rLine %d",lnum);
		ilen = strlen(iline);
		if((ilen > 0) && (iline[0] != ';')) {

			for(i=0; i < 2; i++)
				p[i] = 0;

			sscanf(iline,
			"%1s%1s%1s%1s%1s%x%d%d%d%d%d%d%d%d%d%d%d%d",
			tseq,pczip,trace,pch,pcl,&ntseq,&ttrace,&alurs,&trapon,
			&nodtrap,&f_oreg,&f_left,&f_right,&f_alu,&f_jstat,
			&f_opc,&spare1,&spare2);

			field(ntseq,	0,	0,	16);
			field(ttrace,	0,	4,	2);
			field(alurs,	0,	5,	2);
			field(trapon,	0,	6,	2);
			field(nodtrap,	0,	7,	2);
			field(f_oreg,	1,	0,	2);
			field(f_left,	1,	1,	2);
			field(f_right,	1,	2,	2);
			field(f_alu,	1,	3,	2);
			field(f_jstat,	1,	4,	2);
			field(f_opc,	1,	5,	2);
			field(spare1,	1,	6,	2);
			field(spare2,	1,	7,	2);

			jval = jmask = 0;
			afield(tseq[0],		7,	0x780,	16);
			afield(pczip[0],	6,	0x40,	2);
			afield(trace[0],	5,	0x20, 	2);
			afield(pch[0],		4,	0x10,	2);
			afield(pcl[0],		0,	0x0f,	16);

			for(i=0; i < 2048; i++) {
				aa = (jmask & i) | (~jmask & jval);
				for(j=0; j<2; j++)
					prom[j][aa] = p[j];
			}
		}
	}
	fclose(f);
	printf("\r*** Done reading source file....\n");

	for(i=0; i<2; i++) {
		fname_v[3] = i | '0';
		printf("*** Writing file - %s\n",fname_v);
		f = fopen(fname_v,"w");
		fprintf(f,"$DATA_HEADER$\n$TYPE$\nROM\n$FORMAT$\n");
		fprintf(f,"ADDRESS_DATA\n$BASE$\nD B\n$TOTAL_COLUMNS$");
		fprintf(f,"\n4 8\n$END$\n");
		for(j=0; j<2048; j++) {
			fprintf(f,"%04d ",j);
			for(k=0; k<8; k++) 
				fputc('0' | (1 & (prom[i][j] >> (7 - k))),f);
			fputs("\n",f);
		}
		fclose(f);

		fname_d[3] = i | '0';
		printf("*** Writing file - %s\n",fname_d);
		f = fopen(fname_d,"w");
		fputc(2,f);
		sum = 0;
		for(j=0; j<2048; j++) {
			hex2(f,prom[i][j]);
			fputc(' ',f);
			sum += (0xff & prom[i][j]);
		}
		hex2(f,prominit[i]);
		sum += (0xff & prominit[i]);
		fputc(' ',f);
		fputc(3,f);

		fprintf(f,"              \nCHECKSUM=%04x\n",(sum & 0xffff));
		
		fclose(f);
	}
	printf("*** Done ...\n");
}

int field(n, byte, bit, max) 
int n,byte,bit,max;
{
	if((n >= max) || (n < 0))
		printf(" - Illegal data in field !!!\n");
	p[byte] |= n << bit;
}

afield(n, bit, mask, max)
int n,bit,mask,max;
{
	int i,f;
	if(n == 'X')
		jmask |= mask;
	else {
		f = -1;
		for(i=0; i<max; i++) {
			if(xtab[i] == n)
				f = i;
		}
		if(f == -1) printf(" - Bad character in address field\n");
		jval |= f << bit;
	}
}

int hex2(f, n)
FILE *f;
int n;
{
	fputc(xtab[15 & (n >> 4)],f);
	fputc(xtab[15 & n],f);
}
